(*********************************************************************)
(*                         Diy                                       *)
(*                                                                   *)
(*   Jade Alglave, Luc Maranget INRIA Paris-Rocquencourt, France.    *)
(*                                                                   *)
(*  Copyright 2010 Institut National de Recherche en Informatique et *)
(*  en Automatique. All rights reserved. This file is distributed    *)
(*  under the terms of the Lesser GNU General Public License.        *)
(*********************************************************************)

open Code
open Printf
open Config

module type Conf = sig
  val hout : Hint.out
  val unicond : bool
end

module Make (Co:Conf) (C:XXXCompile.S) : sig

  module A : Arch.S

  type test
  type edge = C.E.edge

(* Returns resolved edges of test *)
  val extract_edges : test -> edge list

(* Build up test, test structure includes
   name & comment given as first two arguments,
   third argument is the last minute check *)
  type info = (string * string) list

  type check = edge list list -> bool
  val make_test :
      string -> ?com:string -> ?info:info -> ?check:check -> edge list -> test
(* Build test from cycle *)
  val test_of_cycle :
      string -> ?com:string -> ?info:info -> ?check:check -> edge list ->
       C.C.node -> test  
(* Dump the given test *)
  val dump_test : test -> unit
  val dump_test_channel : out_channel -> test -> unit


(* Combination of make_test/dump_test *)
(*
  val compile :
      string -> ?com:string -> ?info:MiscParser.info ->
        ?check:check -> edge list -> unit
*)

end
= struct

(* Config *)
  let nprocs = !Config.nprocs
  let eprocs = !Config.eprocs
  let do_observers = !Config.do_observers
  let obs_type = !Config.obs_type
  let optcond = !Config.optcond
  let overload = !Config.overload
  let fno = !Config.fno
  let poll = !Config.poll


  module A = C.A
  module Run = Run.Make(C)

  open C.E
  type edge = C.E.edge
  open C.C

  type info = (string * string) list
  type final =
    | Exists of A.final
    | Forall of A.final list

  type test =
      {
       name : string ;
       com : string ;
       info : info ;
       edges : edge list ;
       init : A.init ;
       prog : A.pseudo list list ;
       final : final ;
     }

  let extract_edges {edges=es; _} = es

(********)
(* Util *)
(********)

module StringSet = MySet.Make(String)

(***************)
(* Compilation *)
(***************)


  let show_in_cond =
    if optcond then
      let valid_edge e = match e.C.E.edge with
      | Rf _ | RfStar _| Fr _ | Ws _ | Hat -> true
      | Po _ | Fenced _ | Dp _|Rmw -> false in
      (fun n -> valid_edge n.prev.edge || valid_edge n.edge)
    else
      (fun _ -> true)

  let add_final_v p r v finals = (A.Reg (p,r),v)::finals

  let add_final p o n finals = match o with
  | Some r ->
      let m,fs = finals in
      if show_in_cond n then
        EventMap.add n.evt (A.Reg (p,r)) m,
        add_final_v p r n.evt.v fs
      else finals
  | None -> finals

  let rec emit_overload st p init ov loc =
    if ov <= 0 then init,[],st
    else
      let loc_ov = sprintf "%s%i" loc ov in
      let _,init,i,st = C.emit_load st p init loc_ov in
      let init,is,st = emit_overload st p init (ov-1) loc in
      init,i@is,st

  let insert_overload n = match n.edge.C.E.edge with
  | Po (_,Dir R,Dir (W|R)) -> true
  | _ -> false

  type prev_load =
    | No       (* Non-existent or irrelevant *)
    | Yes of C.E.dp * C.A.arch_reg


(* Encodes load of first non-initial vakue in chain,
   can poll on value in place of checking it *)
  let do_poll n = match poll,n.prev.edge.C.E.edge,n.evt.v with
  | true,Rf Ext,1 -> true
  | _,_,_ -> false

  let emit_access ro_prev st p init n =
    let init,ip,st = match overload with
    | Some ov  when insert_overload n ->
        emit_overload st p init ov n.evt.loc
    | _ -> init,[],st in
    let o,init,i,st = match ro_prev with
    | No ->
        begin match
          fno,n.evt.dir,n.prev.prev.edge.C.E.edge,n.prev.edge.C.E.edge,
          n.edge.C.E.edge
        with
        | (FnoPrev|FnoBoth|FnoOnly),R,_,Rf _, Fenced _->
            let r,init,i,st = C.emit_fno st p init n.evt.loc in
            Some r,init,i,st
        | (FnoBoth|FnoOnly),R,Rf _, Fenced _,_->
            let r,init,i,st = C.emit_fno st p init n.evt.loc in
            Some r,init,i,st
        | _ ->
            if do_poll n then
              let r,init,i,st =
                C.emit_load_one st p init n.evt.loc in
              Some r,init,i,st
            else
              C.emit_access st p init n.evt
        end
    | Yes (dp,r1) -> C.emit_access_dep st p init n.evt dp r1 in    
    o,init,ip@i,st

let edge_to_prev_load o e = match o with
| None -> No
| Some r ->
    begin match e.C.E.edge with
    | Dp (dp,_,_) -> Yes (dp,r)
    | _ -> No
    end

let get_fence n = match fno,n.edge.C.E.edge with
| FnoOnly,Fenced (fe,_,_,_) ->
    begin match n.evt.dir,n.next.evt.dir with
    | R,R -> None
    | _,_ -> Some fe
    end
| _, Fenced (fe,_,_,_) ->  Some fe
| _,_ -> None

let rec compile_proc loc_writes st p ro_prev init ns = match ns with
| [] -> init,[],(EventMap.empty,[]),st
| n::ns ->
    begin match
      fno,n.evt.dir,n.prev.edge.C.E.edge,n.next.evt.dir,ns with
    | FnoAll,R,_,_,_ ->
        begin match  ro_prev with
        | No -> ()
        | Yes _ -> Warn.fatal "Dependency to fno"
        end ;
        let o,init,i,st = C.emit_fno st p init n.evt.loc in
        let init,is,finals,st =
          compile_proc loc_writes
            st p (edge_to_prev_load (Some o) n.edge)
            init ns in
        init,
        i@(match get_fence n with Some fe -> C.emit_fence fe::is  | _ -> is),
        (if StringSet.mem n.evt.loc loc_writes then
          add_final p (Some o) n finals
        else finals),
        st
    | (FnoRf,R,Rf _,d,m::ns)
    | (_,R,RfStar _,d,m::ns) ->
        let o1,init,i1,lab,st = C.emit_open_fno st p init n.evt.loc in
        let o2,init,i2,st =
          emit_access (edge_to_prev_load (Some o1) n.edge) st p init m in
        let init,i3,st = C.emit_close_fno st p init lab o1 n.evt.loc in
        let init,is,finals,st =
          compile_proc loc_writes
            st p (edge_to_prev_load o2 m.edge)
            init ns in
        init,
        i1@
        (match  get_fence n  with Some fe -> [C.emit_fence fe] | _ -> [])@
        (match d with R -> i2@i3@is | W -> i3@i2@is),
        add_final p (Some o1) n (add_final p o2 m finals),
        st
    | _ ->
        let o,init,i,st = emit_access ro_prev st p init n in
        let init,is,finals,st =
          compile_proc loc_writes
            st p (edge_to_prev_load o n.edge)
            init ns in
        init,
        i@(match get_fence n with Some fe -> C.emit_fence fe::is  | _ -> is),
        (if
          StringSet.mem n.evt.loc loc_writes && not (do_poll n)
        then
          add_final p o n finals
        else finals),
        st
    end

let rec straight_observer st p i x = function
  | [] -> i,[],[]
  | [v] ->
      let r,i,c,_st = C.emit_load st p i x in
      i,c,add_final_v p r v []
  | v::vs ->
      let r,i,c,st = C.emit_load st p i x in
      let i,cs,fs = straight_observer st p i x vs in
      i,c@cs,add_final_v p r v fs


let rec fenced_observer st p i x = function
  | [] -> i,[],[]
  | [v] ->
      let r,i,c,_st = C.emit_load st p i x in
      i,c,add_final_v p r v []
  | v::vs ->
      let r,i,c,st = C.emit_load st p i x in
      let f = [C.emit_fence C.stronger_fence] in
      let i,cs,fs = fenced_observer st p i x vs in
      i,c@f@cs,add_final_v p r v fs

let loop_observer st p i x = function
  | []|[_] -> i,[],[]        
  | v::vs ->
      let r,i,c,st = C.emit_load_not_zero st p i x in
      let rec do_loop st i prev_r = function
        | [] ->  i,[],[]
        | v::vs ->
            let r,i,c,st = C.emit_load_not_eq st p i x prev_r in
            let i,cs,fs = do_loop st i r vs in
            i,c@cs,add_final_v p r v fs in
      let i,cs,fs = do_loop st i r vs in
      i,c@cs,add_final_v p r v fs

let build_observer = match obs_type with
| Straight ->  straight_observer
| Config.Fenced -> fenced_observer
| Loop -> loop_observer

let rec build_observers p i x arg = match arg,do_observers with
  | [],_ -> i,[],[]
  | []::vss,_
  | [_]::vss,(Avoid|Accept) ->
      build_observers p i x vss
  | _,Avoid  -> Warn.fatal "Observer"
  | vs::vss,_ ->
      let i,c,f = build_observer A.st0 p i x vs in
      let i,cs,fs = build_observers (p+1) i x vss in
      i,c::cs,f@fs

let rec check_rec p i = function
  | [] -> i,[],[]
  | (x,vs)::xvs ->
      let i,c,f = match vs with
      | []|[[_]] -> i,[],[]
      | [[_;v]] ->
          begin match do_observers with
          | Local -> i,[],[]
          | Avoid|Accept -> i,[],[A.Loc x,v]
          | Enforce ->  build_observers p i x vs
          end
      | _ ->
          begin match do_observers with
          | Local -> i,[],[]
          | _ -> build_observers p i x vs 
          end in
      let i,cs,fs =
        check_rec (p+List.length c) i xvs in
      i,c@cs,f@fs

  let check_writes p i cos = check_rec p i cos

  let comp_loc_writes n0 =
    let rec do_rec n =
      let k =
        if n.next == n0 then StringSet.empty
        else do_rec n.next in
      match n.evt.dir with
      | W -> StringSet.add n.evt.loc k
      | R -> k in
    do_rec n0

(* Prefetch hints *)

  let do_prefetch_fst proc ns k = match ns with
  | [] -> assert false
  | [n] ->
      begin match  n.prev.edge.C.E.edge,n.edge.C.E.edge with
      | Ws _,Rf _ -> sprintf "%i:%s=F" proc n.evt.loc::k
      | Fr _,Rf _ -> sprintf "%i:%s=T" proc n.evt.loc::k
      | _,_ -> k
      end
  | n::_ ->
      begin match n.prev.edge.C.E.edge with
      | Ws _ ->
          sprintf "%i:%s=F" proc n.evt.loc::k
      | _ -> k
      end


 let do_prefetch_po_lst proc ns k = match ns with
 | []|[_] -> k
 | _ ->
    let n = Misc.last ns in
    begin match n.evt.dir with
    | W ->  sprintf "%i:%s=W" proc n.evt.loc::k
    | R ->  sprintf "%i:%s=T" proc n.evt.loc::k
    end

  let do_prefetch_po_fst proc ns k = match ns with
  | [] -> assert false
  | [_] -> do_prefetch_fst proc ns k
  | n::_ -> sprintf "%i:%s=F" proc n.evt.loc::k

  let do_prefetch_po proc ns k =
    do_prefetch_po_fst proc ns (do_prefetch_po_lst proc ns k)

  let compile_prefetch_po nss =
    let rec do_rec proc = function
      | [] -> []
      | ns::rem ->
          do_prefetch_po proc ns (do_rec (proc+1) rem) in
    let xs = do_rec 0 nss in
    String.concat "," xs


(******************)
(* Affinity hints *)
(******************)

(*  Most of placement computation is now by litmus *)

 let last_edge ns =
   let n = Misc.last ns in n.C.C.edge.C.E.edge

let compile_coms nss =
  List.map
    (fun ns ->
      let open C.E in
      match last_edge ns with
      | Fr _ -> "Fr"
      | Rf _|RfStar _ -> "Rf"
      | Ws _ -> "Ws"
      | _ -> assert false)
      nss


(* Local check of coherence *)

  let do_add_load st p i f x v =
    let r,i,c,st = C.emit_load st p i x in
    i,c,add_final_v p r v f,st

  let do_add_loop st p i f x v w =
    let r,i,c,st = C.emit_load_not_value st p i x v in
    i,c,add_final_v p r w f,st


  let flatten_co cos = List.map (fun (x,vss) -> x,List.concat vss) cos

  let is_load_init e = e.evt.dir = R && e.evt.v = 0

  let add_co_local_check _co ns st p i code f =
    let lst = Misc.last ns in
    match lst.edge.C.E.edge with
    | Ws _|Fr _ when not (is_load_init lst) ->
        let x = lst.evt.loc and v = lst.next.evt.v in
        begin match obs_type with
        | Straight ->
            let i,c,f,st = do_add_load st p i f x v in
            i,code@c,f,st
        | Config.Fenced ->
            let i,c,f,st = do_add_load st p i f x v in
            let c = C.emit_fence C.stronger_fence::c in
           i,code@c,f,st
        | Loop ->
            let i,c,f,st = do_add_loop st p i f x lst.evt.v v in
            i,code@c,f,st
        end
    | _ ->  i,code,f,st

  let compile_cycle ok n =
    Label.reset () ;
    let cos = C.C.coherence n in    
    let co = flatten_co cos in
    if !Misc.verbose > 1 then begin
      eprintf "COHERENCE: " ;
      Misc.pp_list stderr ""
        (fun chan (x,vs) ->
          fprintf chan "<%s:%a>" x
            (fun chan ->
              Misc.pp_list chan "|"
                (fun chan ->
                  Misc.pp_list chan ","
                    (fun chan -> fprintf chan "%i")))
            vs)
        cos ;
      eprintf "\n%!"
    end ;
    let loc_writes = comp_loc_writes n in
    let rec do_rec p i = function
      | [] -> List.rev i,[],(EventMap.empty,[])
      | n::ns ->
          let i,c,(m,f),st = compile_proc loc_writes A.st0 p No i n in
          let i,c,f,st =
            if Co.unicond then  i,c,f,st
            else match do_observers with
            | Local -> add_co_local_check co n st p i c f
            | Avoid|Accept|Enforce -> i,c,f,st in
          let i,c,_ = C.postlude st p i c in
          let i,cs,(ms,fs) = do_rec (p+1) i ns in
          i,c::cs,(union_map m ms,f@fs) in

    let i,c,f =
      if Co.unicond then [],[],[] else check_writes 0 [] cos in
    let splitted =  C.C.split_procs n in
    match splitted with
    | [] -> Warn.fatal "No proc"
    | [_] when not Co.unicond -> Warn.fatal "One proc"
    | _ ->
        let i,c,(m,f) =
          if
            let len =  List.length splitted in
            nprocs <= 0 ||
            (if eprocs then len = nprocs else len <= nprocs)
          then
            let ess = List.map (List.map (fun n -> n.C.C.edge)) splitted in
            if ok ess then
              let i,cs,(m,fs) = do_rec (List.length c) i splitted in
              if
                List.exists
                  (fun (_,loc) -> (loc:string) = Code.ok)
                  i
              then
                (C.A.Loc Code.ok,"1")::i,c@cs,
                (m,(C.A.Loc Code.ok,1)::f@fs)
              else
                i,c@cs,(m,f@fs)
            else Warn.fatal "Last minute check"
          else  Warn.fatal "Too many procs" in
        let r =
          if Co.unicond then
            let evts =
              List.map
                (List.map (fun n -> n.evt))
                splitted in
            let f = Run.run evts m in
            i,c,Forall f
          else
            i,c,Exists f in
        r,(compile_prefetch_po splitted,compile_coms splitted)


(********)
(* Dump *)
(********)

let get_proc = function
  | A.Loc _ -> -1
  | A.Reg (p,_) -> p

let dump_init chan inits =
  fprintf chan "{" ;
  let rec p_rec q = function
    | [] -> fprintf chan "\n}\n"
    | (left,loc)::rem ->
        let p = get_proc left in
        if p <> q then fprintf chan "\n" else fprintf chan " " ;
        fprintf chan "%s=%s;" (A.pp_location left) loc ;
        p_rec p rem in
  p_rec (-1) inits


let rec dump_pseudo = function
  | [] -> []
  | A.Instruction ins::rem -> A.dump_instruction ins::dump_pseudo rem
  | A.Label (lbl,ins)::rem ->
      sprintf "%s:" lbl::dump_pseudo (ins::rem)
  | A.Nop::rem -> dump_pseudo rem
  | A.Macro (m,args)::rem ->
      sprintf "%s(%s)"
        m
        (String.concat ","
           (List.map A.pp_reg args))::
      dump_pseudo rem

let fmt_cols =
  let rec fmt_col p k = function
    | [] -> k
    | cs::prog ->
        (sprintf "P%i" p::dump_pseudo cs)::
        fmt_col (p+1) k prog in
  fmt_col 0 []

  let dump_code chan code =
    let pp = fmt_cols code in
    Misc.pp_prog chan pp

  let dump_state fs =
    String.concat " /\\ " 
      (List.map
         (fun (r,v) -> sprintf "%s=%i" (A.pp_location r) v)
         fs)

  let dump_final chan tst = match tst.final with
  | Exists fs ->
      fprintf chan "%sexists\n" (if !Config.neg then "~" else "") ;
      fprintf chan "(%s)\n" (dump_state fs)
  | Forall ffs ->
      fprintf chan "forall\n" ;
      fprintf chan "%s\n" (Run.dump_cond ffs)
      

  let dump_test_channel chan t =
    fprintf chan "%s %s\n" (Misc.pp_arch A.arch) t.name ;
    if t.com <>  "" then fprintf chan "\"%s\"\n" t.com ;
    List.iter
      (fun (k,v) -> fprintf chan "%s=%s\n" k v)
      t.info ;
    Hint.dump Co.hout t.name t.info ;
    dump_init chan t.init ;
    dump_code chan t.prog ;
    dump_final chan t ;
    ()

let dump_test ({ name = name; _ } as t) =
  let fname = name ^ ".litmus" in
  Misc.output_protect
    (fun chan -> dump_test_channel chan t)
    fname


let test_of_cycle name ?com ?(info=[]) ?(check=(fun _ -> true)) es c =
  let com = match com with None -> pp_edges es | Some com -> com in
  let(init,prog,final),(prf,coms) = compile_cycle check c in
  let coms = String.concat " " coms in
  let info = info@["Prefetch",prf ; "Com",coms; "Orig",com; ] in
  { name=name ; info=info; com=com ;  edges = es ;
    init=init ; prog=prog ; final=final ; }
    
let make_test name ?com ?info ?check es =
  try
    if !Misc.verbose > 1 then eprintf "**Test %s**\n" name ;
    let es,c = C.C.make es in
    test_of_cycle name ?com ?info ?check es c
  with
  | Misc.Fatal msg ->
      Warn.fatal "Test %s [%s] failed:\n%s" name (pp_edges es) msg
  

 type check = edge list list -> bool


end
